﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2014 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Core;
using HeuristicLab.Encodings.SymbolicExpressionTreeEncoding;
using HeuristicLab.EvolutionTracking;

namespace HeuristicLab.Problems.DataAnalysis.Symbolic {
  public class SymbolicDataAnalysisExpressionAfterManipulatorOperator : AfterManipulatorOperator<ISymbolicExpressionTree> {
    public override IOperation Apply() {
      var childVertex = (IGenealogyGraphNode<ISymbolicExpressionTree>)GenealogyGraph.GetByContent(ChildParameter.ActualValue);
      var nodesBefore = (List<ISymbolicExpressionTreeNode>)childVertex.InArcs.First().Data;
      var nodesAfter = ChildParameter.ActualValue.IterateNodesPrefix().ToList();
      IFragment<ISymbolicExpressionTreeNode> fragment = null;

      // try to find if a subtree was replaced by comparing references
      for (int i = 0; i < Math.Min(nodesBefore.Count, nodesAfter.Count); ++i) {
        if (nodesAfter[i] != nodesBefore[i]) {
          fragment = new Fragment<ISymbolicExpressionTreeNode> {
            Root = nodesAfter[i],
            Index1 = i,
            Index2 = i
          };
        }
      }
      // if the fragment could not be identified by reference comparison, we try to identify it by comparing node labels
      // the fragment root will be the lowest common ancestor of all the differing nodes
      if (fragment == null) {
        var list = new List<ISymbolicExpressionTreeNode>();
        var root = ChildParameter.ActualValue.Root;

        for (int i = 0; i < Math.Min(nodesBefore.Count, nodesAfter.Count); ++i) {
          var a = nodesAfter[i];
          var b = nodesBefore[i];

          if (!Equals(a.ToString(), b.ToString())) // label comparison is sufficient
            list.Add(a);
        }

        var lca = LowestCommonAncestor(root, list);
        int index = nodesAfter.IndexOf(lca);
        fragment = new Fragment<ISymbolicExpressionTreeNode> {
          Root = lca,
          Index1 = index,
          Index2 = index
        };
      }

      if (fragment == null)
        throw new ArgumentNullException("Could not identify fragment.");

      childVertex.InArcs.First().Data = fragment;
      return base.Apply();
    }

    private static ISymbolicExpressionTreeNode LowestCommonAncestor(ISymbolicExpressionTreeNode root, List<ISymbolicExpressionTreeNode> nodes) {
      if (nodes.Count == 0)
        throw new ArgumentException("The nodes list should contain at least one element.");

      if (nodes.Count == 1)
        return nodes[0];

      int lowestLevel = nodes.Min(x => root.GetBranchLevel(x));

      // bring the nodes in the nodes to the same level (relative to the root)
      for (int i = 0; i < nodes.Count; ++i) {
        var node = nodes[i];
        var level = root.GetBranchLevel(node);
        for (int j = lowestLevel; j < level; ++j)
          node = node.Parent;
        nodes[i] = node;
      }

      // while not all the elements in the nodes are equal, go one level up
      while (nodes.Any(x => x != nodes[0])) {
        for (int i = 0; i < nodes.Count; ++i)
          nodes[i] = nodes[i].Parent;
      }

      return nodes[0];
    }
  }
}
